import numpy as np
import matplotlib.pyplot as plt
import sklearn.datasets

class Variable(): # for all inputs/outputs in functions we will use Variable (W, b, x, y_prim)
    def __init__(self, value):
        self.value = value # W
        self.grad = np.zeros_like(value) # d_W

# f(x, W, b) = W * x + b (affine transformation)
class LayerLinear():
    # W single scalar => W that can have many values
    # x single scalar => x that can have many values (in_features, out_fetures) => Floors of the house, year built, number of rooms
    # y single scalar => y that can have many values
    def __init__(self, in_features, out_fetures):
        self.W = Variable(
            value=np.random.random((out_fetures, in_features))
        )
        self.b = Variable(
            value=np.zeros(( out_fetures, ))
        )
        self.x: Variable = None
        self.output: Variable = None

    def forward(self, x):
        self.x: Variable = x
        self.output = Variable(
            np.matmul(self.W.value, x.value.transpose()).transpose() + self.b.value
        )
        return self.output

    def backward(self):
        # d_b = 1 * chain_rule_grad
        self.b.grad = 1 * self.output.grad # (batch, out_features)
        # self.b.value => (out_features, )

        # d_W = x * chain_rule_grad
        self.W.grad = np.matmul(
            np.expand_dims(self.output.grad, axis=2),
            np.expand_dims(self.x.value, axis=1))

        # d_x = W * chain_rule_grad
        self.x.grad = np.matmul(
            self.W.value.transpose(),
            self.output.grad.transpose()
        ).transpose()


class LossMSE():
    def __init__(self):
        self.y  = None
        self.y_prim  = None

    def forward(self, y: Variable, y_prim: Variable):
        self.y = y
        self.y_prim = y_prim
        loss = np.mean((y.value - y_prim.value)**2)
        return loss

    def backward(self):
        self.y_prim.grad = -2*(self.y.value - self.y_prim.value)

class LossMAE():
    def __init__(self):
        self.y = None
        self.y_prim = None

    def forward(self, y: Variable, y_prim: Variable):
        self.y = y
        self.y_prim = y_prim
        loss = np.mean(np.abs(y_prim.value - y.value))
        return loss

    def backward(self):
         # (self.y.value - self.y_prim.value) >= 0 => [1, 0, 1, 0] => [1, -1...

        self.y_prim.grad = (((self.y_prim.value - self.y.value) >= 0) - 0.5) * 2.0 # -1.0 or 1.0
        #np.where()

# Sigmoid function 1 / (1 + e^-x)
class LayerSigmoid():
    def __init__(self):
        self.x = None
        self.output = None

    def forward(self, x: Variable):
        self.x = x
        self.output = Variable(1.0 / (1 + np.exp(-x.value)))
        return self.output

    def backward(self):
        self.x.grad = (np.exp(-self.x.value) / (1 + np.exp(-self.x.value)) ** 2) * self.output.grad


class LayerReLU():
    def __init__(self):
        self.x = None
        self.output = None

    def forward(self, x: Variable):
        self.x = x
        self.output = Variable((x.value >= 0) * x.value)
        # [1, -1, -1] >= 0
        # [True, False, False] * 1.0 => [1, 0, 0]
        return self.output

    def backward(self):
        self.x.grad = (self.x.value >= 0) * self.output.grad

class Model:
    def __init__(self):
        self.layers = [
            LayerLinear(in_features=8, out_fetures=4), # W_1*x + b < dW , db
            LayerReLU(),
            LayerLinear(in_features=4, out_fetures=4), # W_2*x + b
            LayerReLU(),
            LayerLinear(in_features=4, out_fetures=1), # W_3*x + b
        ]

    def forward(self, x):
        out = x
        for layer in self.layers:
            out = layer.forward(out)
        return out

    def backward(self):
        for layer in reversed(self.layers):
            layer.backward()

    def parameters(self):
        variables = []
        for layer in self.layers:
            if isinstance(layer, LayerLinear):
                variables.append(layer.W)
                variables.append(layer.b)
        return variables

class OptimizerSGD:
    def __init__(self, parameters, learning_rate):
        self.parameters = parameters
        self.learning_rate = learning_rate

    def step(self):
        for param in self.parameters:
            # W (8, 4)            dW (16, 8, 4)  => (Batch, InFeatures, OutFeatures)
            param.value -= np.mean(param.grad, axis=0) * self.learning_rate  # W = W - dW * alpha


model = Model()
optimizer = OptimizerSGD(
    model.parameters(),
    learning_rate=1e-8
)
loss_fn = LossMAE()

X, Y = sklearn.datasets.load_wine(
    return_X_y=True
)

# shuffle
dataset_len = len(Y)
idxes_rand = np.random.permutation(dataset_len)
X = X[idxes_rand]
Y = Y[idxes_rand]


BATCH_SIZE = 16

for epoch in range(10):
    losses = []
    for idx in range(0, dataset_len-BATCH_SIZE, BATCH_SIZE):
        x = X[idx:idx+BATCH_SIZE]
        y = Y[idx:idx+BATCH_SIZE]

        y_prim = model.forward(Variable(value=x))
        loss = loss_fn.forward(Variable(value=y), y_prim)

        losses.append(loss)

        loss_fn.backward()
        model.backward()

        optimizer.step()

    print(f'loss: {np.mean(losses)}')



